@@ -6,10 +6,9 @@ from threading import Semaphore, Thread, main_thread
 from typing import Any
 from uuid import uuid1
 
-from webview import (FOLDER_DIALOG, OPEN_DIALOG, SAVE_DIALOG, _settings, settings,
+from webview import (FOLDER_DIALOG, OPEN_DIALOG, SAVE_DIALOG, _state, settings,
                      parse_file_type, windows)
 from webview.dom import _dnd_state
-from webview.js.css import disable_text_select
 from webview.menu import Menu, MenuAction, MenuSeparator
 from webview.screen import Screen
 from webview.util import DEFAULT_HTML, create_cookie, js_bridge_call, inject_pywebview
@@ -34,7 +33,6 @@ except ValueError:
 from gi.repository import Gdk, Gio
 from gi.repository import GLib as glib
 from gi.repository import Gtk as gtk
-from gi.repository import Soup
 from gi.repository import WebKit2 as webkit
 
 renderer = 'gtkwebkit2'
@@ -69,6 +67,7 @@ class BrowserView:
         self.js_results = {}
 
         self.window = gtk.ApplicationWindow(title=window.title, application=_app)
+        self.pywebview_window.native = self.window
 
         self.shown = window.events.shown
         self.loaded = window.events.loaded
@@ -127,9 +126,8 @@ class BrowserView:
         self.window.connect('configure-event', self.on_window_configure)
 
         self.js_bridge = BrowserView.JSBridge(window)
-        self.text_select = window.text_select
 
-        storage_path = _settings['storage_path'] or os.path.join(os.path.expanduser('~'), '.pywebview')
+        storage_path = _state['storage_path'] or os.path.join(os.path.expanduser('~'), '.pywebview')
 
         if not os.path.exists(storage_path):
             os.makedirs(storage_path)
@@ -137,7 +135,7 @@ class BrowserView:
         web_context = webkit.WebContext.get_default()
         self.cookie_manager = web_context.get_cookie_manager()
 
-        if not _settings['private_mode']:
+        if not _state['private_mode']:
             self.cookie_manager.set_persistent_storage(
                 os.path.join(storage_path, 'cookies'), webkit.CookiePersistentStorage.SQLITE
             )
@@ -152,11 +150,14 @@ class BrowserView:
         self.webview.connect('decide-policy', self.on_navigation)
         self.webview.connect('drag-data-received', self.on_drag_data)
 
+        if settings['IGNORE_SSL_ERRORS']:
+            web_context.set_tls_errors_policy(webkit.TLSErrorsPolicy.IGNORE)
+
         if settings['ALLOW_DOWNLOADS']:
             web_context.connect('download-started', self.on_download_started)
 
         webkit_settings = self.webview.get_settings().props
-        user_agent = settings.get('user_agent') or _settings['user_agent']
+        user_agent = settings.get('user_agent') or _state['user_agent']
         if user_agent:
             webkit_settings.user_agent = user_agent
 
@@ -186,7 +187,7 @@ class BrowserView:
             wvbg.alpha = 0.0
             self.webview.set_background_color(wvbg)
 
-        if _settings['debug']:
+        if _state['debug']:
             webkit_settings.enable_developer_extras = True
 
             if settings['OPEN_DEVTOOLS_IN_DEBUG']:
@@ -194,13 +195,22 @@ class BrowserView:
         else:
             self.webview.connect('context-menu', lambda a, b, c, d: True)  # Disable context menu
 
-        if _settings['private_mode']:
+        if _state['private_mode']:
             webkit_settings.enable_html5_database = False
             webkit_settings.enable_html5_local_storage = False
 
         self.webview.set_opacity(0.0)
         scrolled_window.add(self.webview)
 
+        if _state['icon']:
+            self.window.set_icon_from_file(_state['icon'])
+            self.window.set_default_icon_from_file(_state['icon'])
+
+        self.pywebview_window.events.before_show.set()
+
+        if window.fullscreen:
+            self.toggle_fullscreen()
+
         if window.real_url is not None:
             self.webview.load_uri(window.real_url)
         elif window.html:
@@ -208,9 +218,6 @@ class BrowserView:
         else:
             self.webview.load_html(DEFAULT_HTML, '')
 
-        if window.fullscreen:
-            self.toggle_fullscreen()
-
     def close_window(self, *data):
         should_cancel = self.pywebview_window.events.closing.set()
 
@@ -244,13 +251,19 @@ class BrowserView:
         return False
 
     def on_drag_data(self, widget, drag_context, x, y, data, info, time):
-        if _dnd_state['num_listeners'] > 0 and data.get_text():
-            files = [
-                (os.path.basename(value), value.replace('file://', ''))
-                for value
-                in data.get_text().split('\n')
-                if value.startswith('file://')
-            ]
+        text = data.get_text()
+        if _dnd_state['num_listeners'] > 0 and text:
+            files = []
+            for value in text.split('\n'):
+                value = value.strip()
+                if value.startswith('file://'):
+                    # Handle 'file://' URIs (e.g., `file:///home/user/file.txt`)
+                    path = value.replace('file://', '')
+                    files.append((os.path.basename(path), path))
+                elif value.startswith('/') and os.path.exists(value):
+                    # Handle direct paths (e.g., `/home/user/file.txt`)
+                    files.append((os.path.basename(value), value))
+
             _dnd_state['paths'] += files
 
         return False
@@ -276,7 +289,11 @@ class BrowserView:
 
     def on_js_bridge_call(self, manager, message):
         body = json.loads(message.get_js_value().to_string())
-        js_bridge_call(self.pywebview_window, body['funcName'], body['params'], body['id'])
+
+        if body['funcName'] == '_pywebviewAlert':
+            self.message_box(body['params'])
+        else:
+            js_bridge_call(self.pywebview_window, body['funcName'], body['params'], body['id'])
 
     def on_window_resize(self, window, allocation):
         if allocation.width != self._last_width or allocation.height != self._last_height:
@@ -299,16 +316,7 @@ class BrowserView:
             glib.idle_add(webview.set_opacity, 1.0)
 
         if status == webkit.LoadEvent.FINISHED:
-            if not self.text_select:
-                webview.evaluate_javascript(
-                    script=disable_text_select,
-                    length=len(disable_text_select),
-                    world_name=None,
-                    source_uri=None,
-                    cancellable=None,
-                    callback=None)
-
-            self._set_js_api()
+            inject_pywebview(renderer, self.js_bridge.window)
 
     def on_download_started(self, session, download):
         download.connect('decide-destination', self.on_download_decide_destination)
@@ -513,8 +521,6 @@ class BrowserView:
 
             semaphore.release()
 
-        self.loaded.wait()
-
         cookies = []
         semaphore = Semaphore(0)
         glib.idle_add(_get_cookies)
@@ -523,72 +529,87 @@ class BrowserView:
         return cookies
 
     def get_current_url(self):
-        self.loaded.wait()
         uri = self.webview.get_uri()
         return uri if uri != 'about:blank' else None
 
     def load_url(self, url):
-        self.loaded.clear()
         self.webview.load_uri(url)
 
     def load_html(self, content, base_uri):
-        self.loaded.clear()
         self.webview.load_html(content, base_uri)
 
-    def evaluate_js(self, script):
+    def evaluate_js(self, script, parse_json):
         def _evaluate_js():
-            self.webview.evaluate_javascript(
-                    script=script,
-                    length=len(script),
-                    world_name=None,
-                    source_uri=None,
-                    cancellable=None,
-                    callback=_callback)
+            try:
+                self.webview.evaluate_javascript(
+                        script=script,
+                        length=len(script),
+                        world_name=None,
+                        source_uri=None,
+                        cancellable=None,
+                        callback=_callback)
+            except Exception:
+                logger.exception(e)
+                result_semaphore.release()
+
 
         def _callback(webview, task):
-            value = webview.evaluate_javascript_finish(task)
-            result = value.to_string() if value else None
+            nonlocal result
+            try:
+                s = script
+                value = webview.evaluate_javascript_finish(task)
+                res = self._convert_js_value(value)
 
-            if unique_id in self.js_results:
-                self.js_results[unique_id]['result'] = result
+                if parse_json and res:
+                    try:
+                        result = json.loads(res)
+                    except Exception:
+                        pass
+                else:
+                    result = res
+
+            except Exception as e:
+                logger.exception(e)
 
             result_semaphore.release()
 
-        unique_id = uuid1().hex
         result_semaphore = Semaphore(0)
-        self.js_results[unique_id] = {'semaphore': result_semaphore, 'result': None}
-
-        self.loaded.wait()
+        result = None
         glib.idle_add(_evaluate_js)
         result_semaphore.acquire()
 
-        result = self.js_results[unique_id]['result']
-        result = (
-            None
-            if result == 'undefined' or result == 'null' or result is None
-            else result
-            if result == ''
-            else json.loads(result)
-        )
-
-        del self.js_results[unique_id]
-
         return result
 
-    def _set_js_api(self):
-        def create_bridge():
-            script = inject_pywebview(self.js_bridge.window, 'gtk', uid=self.pywebview_window.uid)
-            self.webview.evaluate_javascript(
-                    script=script,
-                    length=len(script),
-                    world_name=None,
-                    source_uri=None,
-                    cancellable=None,
-                    callback=None)
+    def message_box(self, message):
+        dialog = gtk.MessageDialog(
+            parent=self.window,
+            flags=gtk.DialogFlags.MODAL & gtk.DialogFlags.DESTROY_WITH_PARENT,
+            type=gtk.MessageType.INFO,
+            buttons=gtk.ButtonsType.OK,
+            message_format=message,
+        )
+        dialog.run()
+        dialog.destroy()
 
-            self.loaded.set()
-
-        glib.idle_add(create_bridge)
+    def _convert_js_value(self, js_value):
+        if not js_value or js_value.is_null() or js_value.is_undefined():
+            return None
+        elif js_value.is_boolean():
+            return js_value.to_boolean()
+        elif js_value.is_number():
+            return js_value.to_double()
+        elif js_value.is_string():
+            return js_value.to_string()
+        elif js_value.is_object():
+            js_object = js_value.to_object()
+            python_dict = {}
+            properties = js_object.get_property_names()
+            for prop in properties:
+                python_dict[prop] = self._js_value_to_python(js_object.get_property(prop))
+            return python_dict
+        else:
+            logger.error(f'Unsupported JavaScriptCore.Value type: {js_value}')
+            return js_value.to_string()
 
 
 def setup_app():
@@ -873,11 +894,11 @@ def create_file_dialog(dialog_type, directory, allow_multiple, save_filename, fi
     return file_names[0]
 
 
-def evaluate_js(script, uid):
+def evaluate_js(script, uid, parse_json=True):
     i = BrowserView.instances.get(uid)
 
     if i:
-        return i.evaluate_js(script)
+        return i.evaluate_js(script, parse_json)
 
 
 def get_position(uid):
@@ -921,7 +942,7 @@ def get_screens():
     n = display.get_n_monitors()
     monitors = [Gdk.Display.get_monitor(display, i) for i in range(n)]
     geometries = [Gdk.Monitor.get_geometry(m) for m in monitors]
-    screens = [Screen(geom.width, geom.height, geom) for geom in geometries]
+    screens = [Screen(geom.x, geom.y, geom.width, geom.height, geom) for geom in geometries]
 
     return screens
 
