@@ -12,7 +12,7 @@ from threading import Event, Semaphore
 
 import clr
 
-from webview import FOLDER_DIALOG, OPEN_DIALOG, SAVE_DIALOG, _settings, windows
+from webview import FOLDER_DIALOG, OPEN_DIALOG, SAVE_DIALOG, _state, windows
 from webview.guilib import forced_gui_
 from webview.menu import Menu, MenuAction, MenuSeparator
 from webview.screen import Screen
@@ -181,16 +181,18 @@ class BrowserView:
 
             if window.initial_x is not None and window.initial_y is not None:
                 self.StartPosition = WinForms.FormStartPosition.Manual
-                self.Location = Point(window.initial_x, window.initial_y)
+                self.Location = Point(
+                    int(window.initial_x * self.scale_factor),
+                    int(window.initial_y * self.scale_factor)
+                )
             elif window.screen:
                 self.StartPosition = WinForms.FormStartPosition.Manual
                 x = int(
-                    window.screen.frame.X * self.scale_factor + (window.screen.width - window.initial_width) * self.scale_factor / 2
-                    if window.screen.frame.X >= 0
-                    else window.screen.frame.X * self.scale_factor - window.screen.width * self.scale_factor / 2
+                    window.screen.x * self.scale_factor + (window.screen.width - window.initial_width) * self.scale_factor / 2
+                    if window.screen.x >= 0
+                    else window.screen.X * self.scale_factor + window.screen.width / 2
                 )
-
-                y = int(window.screen.frame.Y * self.scale_factor + (window.screen.height - window.initial_height) * self.scale_factor / 2)
+                y = int(window.screen.y * self.scale_factor + (window.screen.height - window.initial_height) * self.scale_factor / 2)
                 self.Location = Point(x, y)
             else:
                 self.StartPosition = WinForms.FormStartPosition.CenterScreen
@@ -285,6 +287,10 @@ class BrowserView:
             def _shutdown():
                 if is_cef:
                     CEF.shutdown()
+                elif is_chromium:
+                    self.hide()
+                    self.browser.clear_user_data()
+
                 WinForms.Application.Exit()
 
             if not is_cef:
@@ -345,24 +351,9 @@ class BrowserView:
         def on_move(self, sender, args):
             self.pywebview_window.events.moved.set(self.Location.X, self.Location.Y)
 
-        def evaluate_js(self, script):
-            def _evaluate_js():
-                self.browser.evaluate_js(
-                    script, semaphore, js_result
-                ) if is_chromium else self.browser.evaluate_js(script)
-
-            semaphore = Semaphore(0)
-            js_result = []
-
-            self.loaded.wait()
-            self.Invoke(Func[Type](_evaluate_js))
-            semaphore.acquire()
-
-            if is_chromium:
-                result = js_result.pop()
-                return result
-
-            return self.browser.js_result
+        def evaluate_js(self, script, parse_json):
+            result = self.browser.evaluate_js(script, parse_json)
+            return result
 
         def clear_cookies(self):
             def _clear_cookies():
@@ -383,8 +374,6 @@ class BrowserView:
                 logger.error('get_cookies() is not implemented for this platform')
                 return cookies
 
-            self.loaded.wait()
-
             semaphore = Semaphore(0)
 
             self.Invoke(Func[Type](_get_cookies))
@@ -632,14 +621,14 @@ _already_set_up_app = False
 def init_storage():
     global cache_dir
 
-    if not _settings['private_mode'] or _settings['storage_path']:
+    if not _state['private_mode'] or _state['storage_path']:
         try:
             data_folder = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData)
 
             if not os.access(data_folder, os.W_OK):
                 data_folder = Environment.GetFolderPath(Environment.SpecialFolder.UserProfile)
 
-            cache_dir = _settings['storage_path'] or os.path.join(data_folder, 'pywebview')
+            cache_dir = _state['storage_path'] or os.path.join(data_folder, 'pywebview')
 
             if not os.path.exists(cache_dir):
                 os.makedirs(cache_dir)
@@ -815,8 +804,6 @@ def load_url(url, uid):
     if not i:
         return
 
-    i.loaded.clear()
-
     if is_cef:
         CEF.load_url(url, uid)
     else:
@@ -928,13 +915,13 @@ def destroy_window(uid):
         i.browser.js_result_semaphore.release()
 
 
-def evaluate_js(script, uid, result_id=None):
+def evaluate_js(script, uid, parse_json, result_id=None):
     if is_cef:
-        return CEF.evaluate_js(script, result_id, uid)
+        return CEF.evaluate_js(script, result_id, parse_json, uid)
 
     i = BrowserView.instances.get(uid)
     if i:
-        return i.evaluate_js(script)
+        return i.evaluate_js(script, parse_json)
 
 
 def get_position(uid):
@@ -951,7 +938,7 @@ def get_size(uid):
 
 
 def get_screens():
-    screens = [Screen(s.Bounds.Width, s.Bounds.Height, s.WorkingArea) for s in WinForms.Screen.AllScreens]
+    screens = [Screen(s.Bounds.X, s.Bounds.Y, s.Bounds.Width, s.Bounds.Height, s.WorkingArea) for s in WinForms.Screen.AllScreens]
     return screens
 
 
