@@ -13,17 +13,19 @@ import os
 import re
 import sys
 import traceback
+from collections import UserDict
+from glob import glob
 from http.cookies import SimpleCookie
 from platform import architecture
 from threading import Thread
-from typing import TYPE_CHECKING, Any
+from typing import TYPE_CHECKING, Any, Callable
 from uuid import uuid4
 
 import webview
 
-from webview.js import api, dom_json, events, npo, polyfill
 from webview.dom import _dnd_state
 from webview.errors import WebViewException
+import urllib.parse
 
 if TYPE_CHECKING:
     from webview.window import Window
@@ -43,6 +45,29 @@ DEFAULT_HTML = """
 logger = logging.getLogger('pywebview')
 
 
+
+class ImmutableDict(UserDict):
+    """"
+    A dictionary that does not allow adding new keys or deleting existing ones.
+    Only existing keys can be modified.
+    """
+
+    def __init__(self, initial_data=None, **kwargs):
+        self.data = {}
+        if initial_data:
+            self.data.update(initial_data)
+        if kwargs:
+            self.data.update(kwargs)
+
+    def __setitem__(self, key, value):
+        if key not in self.data:
+            raise KeyError(f"Cannot add new key '{key}'. Only existing keys can be modified.")
+        super().__setitem__(key, value)
+
+    def __delitem__(self, key):
+        raise KeyError('Deleting keys is not allowed.')
+
+
 def is_app(url: str | None) -> bool:
     """Returns true if 'url' is a WSGI or ASGI app."""
     return callable(url)
@@ -65,13 +90,14 @@ def get_app_root() -> str:
     if hasattr(sys, '_MEIPASS'):  # Pyinstaller
         return sys._MEIPASS
 
+    if os.getenv('RESOURCEPATH'): # py2app
+        return os.getenv('RESOURCEPATH')
+
     if getattr(sys, 'frozen', False):  # cx_freeze
         return os.path.dirname(sys.executable)
 
-    if 'pytest' in sys.modules and os.getenv('PYTEST_CURRENT_TEST'):
-        test_file = os.getenv('PYTEST_CURRENT_TEST').split('::')[0]
-
-        return os.path.dirname(os.path.join(os.getcwd(), test_file))
+    if 'pytest' in sys.modules and os.getenv('PYWEBVIEW_TEST'):
+        return os.path.join(os.path.dirname(__file__), '..', 'tests')
 
     if hasattr(sys, 'getandroidapilevel'):
         return os.getenv('ANDROID_APP_PATH')
@@ -100,7 +126,7 @@ def base_uri(relative_path: str = '') -> str:
 
 def create_cookie(input_: dict[Any, Any] | str) -> SimpleCookie[str]:
     if isinstance(input_, dict):
-        cookie = SimpleCookie[str]()
+        cookie = SimpleCookie()
         name = input_['name']
         cookie[name] = input_['value']
         cookie[name]['path'] = input_['path']
@@ -133,9 +159,11 @@ def parse_file_type(file_type: str) -> tuple[str, str]:
     raise ValueError(f'{file_type} is not a valid file filter')
 
 
-def inject_pywebview(window: Window, platform: str, uid: str = '') -> str:
+def inject_pywebview(platform: str, window: Window) -> str:
     """"
-    Injects a global window.pywebview object
+    Generates and injects a global window.pywebview object. The object contains exposed API functions
+    as well as utility functions required by pywebview. The function fires before_load event before
+    injecting the object and loaded event after the object is injected.
     """
     exposed_objects = []
 
@@ -143,7 +171,6 @@ def inject_pywebview(window: Window, platform: str, uid: str = '') -> str:
         params = list(inspect.getfullargspec(func).args)
         return params
 
-
     def get_functions(obj: object, base_name: str = '', functions: dict[str, object] = None):
         if obj in exposed_objects:
             return functions
@@ -158,6 +185,7 @@ def inject_pywebview(window: Window, platform: str, uid: str = '') -> str:
 
             if name.startswith('_'):
                 continue
+
             attr = getattr(obj, name)
             if inspect.ismethod(attr):
                 functions[full_name] = get_args(attr)[1:]
@@ -179,37 +207,35 @@ def inject_pywebview(window: Window, platform: str, uid: str = '') -> str:
 
         return [{'func': name, 'params': params} for name, params in functions.items()]
 
-    try:
-        func_list = generate_func()
-    except Exception as e:
-        logger.exception(e)
-        func_list = []
+    def generate_js_object():
+        window.run_js(js_code)
+        window.events._pywebviewready.set()
+        logger.debug('_pywebviewready event fired')
 
-    js_code = (
-        npo.src
-        + polyfill.src
-        + api.src
-        % {
-            'token': _TOKEN,
-            'platform': platform,
-            'uid': uid,
-            'func_list': func_list,
-            'js_api_endpoint': window.js_api_endpoint,
-        }
-        + dom_json.src
-        + events.src
-        % {
-            'drag_selector': webview.DRAG_REGION_SELECTOR,
-            'zoomable': str(window.zoomable).lower(),
-            'draggable': str(window.draggable).lower(),
-            'easy_drag': str(platform == 'chromium' and window.easy_drag and window.frameless).lower(),
-        }
-    )
+        try:
+            with window._expose_lock:
+                func_list = generate_func()
+                window.run_js(finish_script % {
+                    'functions': json.dumps(func_list)
+                })
+                window.events.loaded.set()
+                logger.debug('loaded event fired')
+        except Exception as e:
+            logger.exception(e)
+            window.events.loaded.set()
 
-    return js_code
+    window.events.before_load.set()
+    logger.debug('before_load event fired. injecting pywebview object')
+    js_code, finish_script = load_js_files(window, platform)
+    thread = Thread(target=generate_js_object)
+    thread.start()
 
 
 def js_bridge_call(window: Window, func_name: str, param: Any, value_id: str) -> None:
+    """
+    Calls a function from the JS API and executes it in Python. The function is executed in a separate
+    thread to prevent blocking the UI thread. The result is then passed back to the JS API.
+    """
     def _call():
         try:
             result = func(*func_params)
@@ -246,11 +272,11 @@ def js_bridge_call(window: Window, func_name: str, param: Any, value_id: str) ->
         if event['type'] == 'drop':
             files = event['dataTransfer'].get('files', [])
             for file in files:
-                path = [item for item in _dnd_state['paths'] if item[0] == file['name']]
+                path = [item for item in _dnd_state['paths'] if urllib.parse.unquote(item[0]) == file['name']]
                 if len(path) == 0:
                     continue
 
-                file['pywebviewFullPath'] = path[0][1]
+                file['pywebviewFullPath'] = urllib.parse.unquote(path[0][1])
                 _dnd_state['paths'].remove(path[0])
 
         for handler in element._event_handlers.get(event['type'], []):
@@ -282,11 +308,105 @@ def js_bridge_call(window: Window, func_name: str, param: Any, value_id: str) ->
             thread = Thread(target=_call)
             thread.start()
         except Exception:
-            logger.exception('Error occurred while evaluating function %s', func_name)
+            logger.exception(
+                'Error occurred while evaluating function %s', func_name)
     else:
         logger.error('Function %s() does not exist', func_name)
 
 
+def load_js_files(window: Window, platform: str) -> str:
+    """
+    Load JS files in the order they should be loaded.
+    The order is polyfill, api, the rest and finish.js.
+    Return the concatenated JS code and the finish script, which must be loaded last and
+    separately in order to
+    """
+    js_dir = get_js_dir()
+    logger.debug('Loading JS files from %s', js_dir)
+    js_files = glob(os.path.join(js_dir, '**', '*.js'), recursive=True)
+    ordered_js_files = sort_js_files(js_files)
+    js_code = ''
+    finish_script = ''
+
+    for file in ordered_js_files:
+        with open(file, 'r') as f:
+            name = os.path.splitext(os.path.basename(file))[0]
+            content = f.read()
+            params = {}
+
+            if name == 'api':
+                params = {
+                    'token': _TOKEN,
+                    'platform': platform,
+                    'uid': window.uid,
+                    'js_api_endpoint': window.js_api_endpoint
+                }
+            elif name == 'customize':
+                params = {
+                    'text_select': str(window.text_select),
+                    'drag_selector': webview.DRAG_REGION_SELECTOR,
+                    'zoomable': str(window.zoomable),
+                    'draggable': str(window.draggable),
+                    'easy_drag': str(platform == 'chromium' and window.easy_drag and window.frameless).lower(),
+                }
+            elif name == 'finish':
+                finish_script = content
+                continue
+            elif name == 'polyfill' and platform != 'mshtml':
+                continue
+
+            js_code += content % params
+
+    return js_code, finish_script
+
+
+def get_js_dir() -> str:
+    """
+    Get the path to the directory with Javascript files.
+    """
+    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'js')
+
+    if os.path.exists(path):
+        return path
+
+    # try py2app frozen path. This is hacky, but it works.
+    # See https://github.com/r0x0r/pywebview/issues/1565
+    if '.zip' in path:
+        base_path = path.split('.zip')[0]
+        dir = os.path.dirname(base_path)
+
+        for file in os.listdir(dir):
+            if file.startswith('python') and os.path.isdir(os.path.join(dir, file)):
+                js_path = os.path.join(dir, file, 'webview', 'js')
+                if os.path.exists(js_path):
+                    return js_path
+
+    raise FileNotFoundError('Cannot find JS directory in %s' % path)
+
+
+def sort_js_files(js_files: list[str]) -> list[str]:
+    """
+    Sorts JS files in the order they should be loaded. Polyfill first, then API, then the rest and
+    finally finish.js that fires a pywebviewready event.
+    """
+    LOAD_ORDER = { 'polyfill': 0, 'api': 1 }
+
+    ordered_js_files = []
+    remaining_js_files = []
+
+    for file in js_files:
+        basename = os.path.splitext(os.path.basename(file))[0]
+        if basename not in LOAD_ORDER:
+            ordered_js_files.append(file)
+        else:
+            remaining_js_files.append((basename, file))
+
+    for basename, file in sorted(remaining_js_files, key=lambda x: LOAD_ORDER[x[0]]):
+        ordered_js_files.insert(LOAD_ORDER[basename], file)
+
+    return ordered_js_files
+
+
 def escape_string(string: str) -> str:
     return (
         string.replace('\\', '\\\\').replace('"', r"\"").replace('\n', r'\n').replace('\r', r'\r')
@@ -389,4 +509,5 @@ def css_to_camel(css_case_string: str) -> str:
 
 
 def android_jar_path() -> str:
-    return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'lib', 'pywebview-android.jar')
\ No newline at end of file
+    return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'lib', 'pywebview-android.jar')
+
